/*
Copyright (c) 2011 NetIO Devices Co.,Ltd
*/
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include "hd44780.h"

#define BAUD 9600
#define DEFAULT_HVG_PULSEWIDTH 70;

#define DEFAULT_GAMMA_SENSITIVITY    150 /* Gamma sensibility of the GM tube */
#define DEFAULT_ALARM_THRESHOLD      10  /* Alarm rings when the dose rate[uSv/H] exceed this value */

#define DETECTION_PULSE_WIDTH 4

volatile uint8_t sgtc;       /* sound gate-on time counter */
volatile uint8_t pgtc;       /* count pluse gate-on time counter */
volatile uint8_t sec_tmr;    /* second timer */
volatile uint8_t cbuf_idx;   /* circular buffer index */
volatile int8_t alarm_on;    /* alarm */
volatile int8_t key0_flg;    /* chattering avoider for key0 */
volatile int8_t key1_flg;    /* chattering avoider for key1 */
volatile int8_t sma_idx;     /* SMA index */
volatile uint16_t cbuf[20];  /* circular buffer */
volatile uint16_t sma[4];    /* Simple Moving Average */
volatile uint32_t cpm;       /* true-CPM counter */
volatile uint32_t total_cnt; /* total detection */
volatile uint16_t cpx;       /* counts per 3 seconds */

volatile uint16_t gamma_sensitivity;
volatile uint16_t alarm_trigger_cpm;
volatile uint16_t hvg_pulsewidth;

volatile char CmdBuf[16];
volatile int8_t cptr;
volatile int8_t sout;

static uint8_t eep_dipsw __attribute__((section(".eeprom"))) = 0x01;
static uint16_t eep_gms __attribute__((section(".eeprom"))) = DEFAULT_GAMMA_SENSITIVITY;
static uint16_t eep_atc __attribute__((section(".eeprom"))) = DEFAULT_GAMMA_SENSITIVITY * DEFAULT_ALARM_THRESHOLD;
static uint16_t eep_hvg __attribute__((section(".eeprom"))) = DEFAULT_HVG_PULSEWIDTH;

const prog_char CR[] = {'G','C','1','0',' ',' ','R','e','v','.','2','F','\r','\n',
                        'N','e','t','I','O',' ','D','e','v','i','c','e','s'};

const prog_char strDisp[] = {0xE4,'S','v',':',' ',' ',' ',' ',' ',' ',' ','T', 'o', 't', 'a', 'l','C','P','M',':',0};

void SerCmdProc(char ch);
void SerCmdExec(void);
void sio_init(uint16_t baud);
void sio_puts(char *buf);
void dispCPM(void);

ISR(USART_RX_vect) {
	if (bit_is_clear(UCSRA, FE)) {
		SerCmdProc(UDR);
	}
}

ISR(TIMER0_OVF_vect)
{
	if (alarm_on)
		PORTB ^= _BV(PB4);

	if (sgtc) {
		if ((sgtc % 2) == 0)
			PORTB ^= _BV(PB4);
		sgtc--;
		if (sgtc == 0)
			PORTB &= ~_BV(PB0);
	}

	if (pgtc) {
		pgtc--;

		if (pgtc == 0)
			PORTB &= ~_BV(PB3);
	}

	if (bit_is_clear(PINA, PA0)) {
		if (key0_flg == 0) {
			int i;
			for (i=0; i<20; i++) cbuf[i] = 0;
			total_cnt = 0;
			lcd_locate(1, 11);
			for (i = 0; i < 5; i++) lcd_putc(' ');
			key0_flg = 1;
		}
	} else {
		key0_flg = 0;
	}

	if (bit_is_clear(PINA, PA1)) {
		if (key1_flg == 0) {
			DDRB ^= _BV(PB4);
			if (bit_is_clear(DDRB, PB4)) {
				eeprom_busy_wait();
				eeprom_write_byte(&eep_dipsw, eep_dipsw & ~0x01);
			} else {
				eeprom_busy_wait();
				eeprom_write_byte(&eep_dipsw, eep_dipsw | 0x01);
			}
			key1_flg = 1;

		}
	} else {
		key1_flg = 0;
	}

}

ISR(TIMER1_COMPA_vect)
{
	int i;

	sec_tmr++;

	if (sec_tmr % 3 == 0) {
		cbuf[cbuf_idx++] = cpx;
		if (cbuf_idx == 20) cbuf_idx = 0;
		cpx = 0;

		cpm = 0;
		for (i = 0; i < 20; i++) cpm += cbuf[i];

		if (cpm > 65535) cpm = 65535;
	}

	if (sec_tmr == 60) sec_tmr = 0;

	dispCPM();
}

ISR(INT0_vect)
{
	cpx++;
	total_cnt++;

	PORTB |= _BV(PB0);
	PORTB |= _BV(PB3);

	sgtc = 80;
	pgtc = DETECTION_PULSE_WIDTH;
}

void SerCmdProc(char ch)
{
	switch(ch) {
		case '\b':
				if (cptr > 0) {
					cptr--;
					sio_puts("\b \b");
				}
				break;
		case '\r':
				if (cptr < 16) {
					CmdBuf[cptr] = 0x00;
					sio_puts("\r\n");
					SerCmdExec();
				}
				cptr = 0;
				break;
		case '\n':
				break;
		default:
				while ( !(UCSRA & (1<<UDRE)) );
				UDR = ch;
				if (cptr < (16 - 1)) {
					CmdBuf[cptr] = ch;
					cptr++;
				} else {
					cptr = 0;
					sio_puts("?\r\n");
				}
				break;
	}
}

void SerCmdExec(void)
{
	int i;
	char* ptr;
	char buf[8];

	if (memcmp((char*)CmdBuf, "stop", 4) == 0) {
		sout = 0;
	} else
	if (memcmp((char*)CmdBuf, "go", 2) == 0) {
		sout = 1;
	} else
	if (memcmp((char*)CmdBuf, "set ", 4) == 0) {
		ptr = (char*)CmdBuf + 4;
		if (memcmp(ptr, "ttc=", 4) == 0) {
			ptr += 4;
			total_cnt = atol(ptr);
			lcd_locate(1, 11);
			for (i = 0; i < 5; i++) lcd_putc(' ');
		} else
		if (memcmp(ptr, "gms=", 4) == 0) {
			ptr += 4;
			gamma_sensitivity = atoi(ptr);
		} else
		if (memcmp(ptr, "atc=", 4) == 0) {
			ptr += 4;
			alarm_trigger_cpm = atoi(ptr);
		} else
		if (memcmp(ptr, "hvg=", 4) == 0) {
			ptr += 4;
			hvg_pulsewidth = atoi(ptr);
			OCR0A = hvg_pulsewidth;
		} else
		if (memcmp(ptr, "snd=", 4) == 0) {
			ptr += 4;
			if (memcmp(ptr, "on", 2) == 0) {
				DDRB |= _BV(PB4);
			} else
			if (memcmp(ptr, "off", 3) == 0) {
				DDRB &= ~_BV(PB4);
			}
		}
	} else
	if (memcmp((char*)CmdBuf, "save", 4) == 0) {
		eeprom_busy_wait();
		eeprom_write_word(&eep_gms, gamma_sensitivity);
		eeprom_busy_wait();
		eeprom_write_word(&eep_atc, alarm_trigger_cpm);
		eeprom_busy_wait();
		eeprom_write_word(&eep_hvg, hvg_pulsewidth);
		if (bit_is_clear(DDRB, PB4)) {
			eeprom_busy_wait();
			eeprom_write_byte(&eep_dipsw, eep_dipsw & ~0x01);
		} else {
			eeprom_busy_wait();
			eeprom_write_byte(&eep_dipsw, eep_dipsw | 0x01);
		}
	} else
	if (memcmp((char*)CmdBuf, "show", 4) == 0) {
			sio_puts("ttc: ");
			ultoa(total_cnt, buf, 10);
			sio_puts(buf);
			sio_puts("\r\n");

			sio_puts("gms: ");
			utoa(gamma_sensitivity, buf, 10);
			sio_puts(buf);
			sio_puts("\r\n");

			sio_puts("atc: ");
			utoa(alarm_trigger_cpm, buf, 10);
			sio_puts(buf);
			sio_puts("\r\n");

			sio_puts("hvg: ");
			utoa(hvg_pulsewidth, buf, 10);
			sio_puts(buf);
			sio_puts("\r\n");
	}
}

void sio_init(uint16_t baud)
{
	UBRRH = (unsigned char)(baud>>8);
	UBRRL = (unsigned char)baud;
	UCSRB = _BV(RXEN)| _BV(TXEN) | _BV(RXCIE);
}

void sio_puts(char *buf)
{
	while (*buf) {
		while ( !(UCSRA & (1<<UDRE)) );
		UDR = *buf++;
	}
}

void dispCPM(void)
{
	int i;
	char buf[8];
	char* ptr;
	uint32_t tmp;

	sma[sma_idx++] = cpm;
	if (sma_idx == 4) sma_idx = 0;

	tmp  = 0;
	for (i = 0; i < 4; i++) tmp += sma[i];
	cpm = ((tmp << 1) / 4 + 1) >> 1;

	if (cpm > alarm_trigger_cpm)
		alarm_on = 1;
	else
		alarm_on = 0;

	lcd_locate(0, 5);
	for (i = 0; i < 6; i++) lcd_putc(' ');
	lcd_locate(0, 5);
	ultoa((cpm * 1000 / gamma_sensitivity), buf, 10);

	if (strlen(buf) > 3) {
		ptr = buf;
		for (i = 0; i < strlen(buf) - 3; i++) {
			lcd_putc(*ptr++);
		}
		lcd_putc('.');
		ptr = buf + strlen(buf) - 3;
	} else {
		lcd_putc('0');
		lcd_putc('.');
		for (i = 0; i < 3 - strlen(buf); i++) {
			lcd_putc('0');
		}
		ptr = buf;
	}
	while (*ptr) lcd_putc(*ptr++);
	lcd_locate(0, 10);
	lcd_putc(' ');
	lcd_putc('T');


	lcd_locate(1, 5);
	for (i = 0; i < 6; i++) lcd_putc(' ');
	lcd_locate(1, 5);
	ultoa(cpm, buf, 10);
	ptr = buf;
	while (*ptr) lcd_putc(*ptr++);

	if (sout) {
		sio_puts(buf);
		sio_puts("\r\n");
	}

	lcd_locate(1, 11);
	for (i = 0; i < 5; i++) lcd_putc(' ');
	lcd_locate(1, 11);
	ultoa(total_cnt % 100000, buf, 10);
	ptr = buf;
	while (*ptr) lcd_putc(*ptr++);
}

int main(void)
{
	int i;

	DDRB |= _BV(PB0) | _BV(PB2) | _BV(PB3);

	PORTA |= _BV(PA0) | _BV(PA1);
	PORTD |= _BV(PD0);

	TCCR0A |= _BV(COM0A1) | _BV(WGM01) | _BV(WGM00);
	TCCR0B |= _BV(CS01);
	OCR0A = 70;

	TCCR1B |= _BV(WGM12) | _BV(CS12) | _BV(CS10);
	OCR1A = 7813;

	MCUCR |= _BV(ISC01) | _BV(ISC00);

	TIMSK |= _BV(OCIE1A) | _BV(TOIE0);

	GIMSK |= _BV(INT0);

	cpx  = 0;
	sgtc = 0;
	pgtc = 0;
	sec_tmr   = 0;
	cbuf_idx  = 0;
	total_cnt = 0;
	alarm_on  = 0;
	key0_flg  = 0;
	key1_flg  = 0;

        cptr = 0;
	sout = 1;

	/* Clear Circular Buffer */
	for (i = 0; i < 20; i++) cbuf[i] = 0;

	/* Clear SMA Buffer */
	for (i = 0; i < 4;  i++) sma[i] = 0;

	eeprom_busy_wait();
	if ( eeprom_read_byte(&eep_dipsw) & 0x01 )
		DDRB |= _BV(PB4);

	eeprom_busy_wait();
	gamma_sensitivity = eeprom_read_word((uint16_t*)&eep_gms);

	eeprom_busy_wait();
	alarm_trigger_cpm = eeprom_read_word((uint16_t*)&eep_atc);

	eeprom_busy_wait();
	hvg_pulsewidth = eeprom_read_word((uint16_t*)&eep_hvg);
	OCR0A = hvg_pulsewidth;

	eeprom_busy_wait();
	OSCCAL = eeprom_read_byte((uint8_t*)0x10);

	sio_init(F_CPU / 16 / BAUD - 1);

	lcd_init();

	for (i = 0; i < 27; i++)
		lcd_putc(pgm_read_byte(&CR[i]));

	sio_puts("#S\r\n");

	_delay_ms(2000);

	lcd_putc('\f');

	for (i = 0; i < 20; i++)
		lcd_putc(pgm_read_byte(&strDisp[i]));

	sei();

	while (1);

}
